=begin

	MetaRuby
	file: ProcAsArray (sample for ArrayMixin usage)

	Copyright (c) 2001 by Mathieu Bouchard
	Licensed under the same license as Ruby.

=end

require 'ArrayMixin'

# usage:
#	UndoableArray.new(length,init)
#	UndoableArray[e0,e1,e2,e3,...]

class UndoableArray
	class<<self; def [](*v); new.replace v; end; end

	attr_reader :history

	def initialize(n=0,v=nil)
		@inner = []
		@history = []
		super
	end

	module Implementation
		def length; @inner.length; end
		def get(i); @inner[i]; end
		def get_seq(i,n); @inner[i,n]; end
		def put(i,v)
#			puts "put(#{i},#{v.inspect})"
			@history << [:put, i, @inner[i]]
			@inner[i]=v
			nil
		end
		def put_seq(i,n,v)
#			puts "put_seq(#{i},#{n},#{v.inspect})"
			@history << [:put_seq, i, v.length, @inner[i,n]]
			@inner[i,n]=v
			nil
		end
	end

	include ArrayMixin
	include Implementation
	include ArrayContract # for safety

	def modify
		history_backup = @history
		@history = []
		begin
			super
		ensure
			history_backup << @history
			@history = history_backup
		end
	end

	def undo
		history_backup = @history
		@history = []
		todo = history_backup.pop
#		print "todo: ";p todo
		self._send_rec(todo)
		@history = history_backup
		self
	end

	def _send_rec(array)
		if Array===array[0] then
			array.reverse_each{|x| _send_rec(x) }
		else
			__send__(*array)
		end
	end

# disabling this feature that does not work yet
#	# defining this to create an instance of Array, not UndoableArray
#	def dup
#		to_a
#	end
end
